# Modified work Copyright (c) 2017-2022 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Fortran type-declaration statements.

"""

__all__ = [
    "Integer",
    "Real",
    "DoublePrecision",
    "Complex",
    "DoubleComplex",
    "Character",
    "Logical",
    "Byte",
    "TypeStmt",
    "Class",
    "intrinsic_type_spec",
    "declaration_type_spec",
    "Implicit",
]

import re
import string
from fparser.common.base_classes import (
    Statement,
    BeginStatement,
    EndStatement,
    AttributeHolder,
    Variable,
)
from fparser.common.utils import (
    split_comma,
    AnalyzeError,
    name_re,
    is_entity_decl,
    is_name,
    parse_array_spec,
)

# Intrinsic type specification statements


class TypeDeclarationStatement(Statement):
    """
    Class capturing various sorts of type declaration.

    ::

        <declaration-type-spec> [ [, <attr-spec>] :: ] <entity-decl-list>
        <declaration-type-spec> = <intrinsic-type-spec>
                                  | TYPE ( <derived-type-spec> )
                                  | CLASS ( <derived-type-spec> )
                                  | CLASS ( * )

        <derived-type-spec> = <type-name> [ ( <type-param-spec-list> ) ]
        <type-param-spec> = [ <keyword> = ] <type-param-value>
        <type-param-value> = <scalar-int-expr> | * | :

        <intrinsic-type-spec> = INTEGER [<kind-selector>]
                                | REAL [<kind-selector>]
                                | DOUBLE PRECISION
                                | COMPLEX [<kind-selector>]
                                | CHARACTER [<char-selector>]
                                | LOGICAL [<kind-selector>]

        <kind-selector> = ( [ KIND = ] <scalar-int-initialization-expr> )
        EXTENSION:
          <kind-selector> = ( [ KIND = ] <scalar-int-initialization-expr> )
                            | * <length>

        <char-selector> = <length-selector>
                          | ( LEN = <type-param-value>,
                              KIND = <scalar-int-initialization-expr> )
                          | ( <type-param-value>, [ KIND = ]
                              <scalar-int-initialization-expr> )
                          | ( KIND = <scalar-int-initialization-expr> [,
                              LEN = <type-param-value>] )
        <length-selector> = ( [ LEN = ] <type-param-value> )
                            | * <char-length> [ , ]
        <char-length> = ( <type-param-value> ) | <scalar-int-literal-constant>

        <attr-spec> = <access-spec> | ALLOCATABLE | ASYNCHRONOUS
                      | DIMENSION ( <array-spec> ) | EXTERNAL
                      | INTENT ( <intent-spec> ) | INTRINSIC
                      | <language-binding-spec> | OPTIONAL
                      | PARAMETER | POINTER | PROTECTED | SAVE
                      | TARGET | VALUE | VOLATILE
        <entity-decl> = <object-name> [ ( <array-spec> ) ] [ * <char-length> ]
                        [ <initialization> ]
                      | <function-name> [ * <char-length> ]
        <initialization> =  = <initialization-expr>
                            | => NULL
        <access-spec> = PUBLIC | PRIVATE
        <language-binding-spec> = BIND ( C [ , NAME =
                                         <scalar-char-initialization-expr>] )
        <array-spec> =   <explicit-shape-spec-list>
                       | <assumed-shape-spec-list>
                       | <deferred-shape-spec-list>
                       | <assumed-size-spec>
        <explicit-shape-spec> = [ <lower-bound> : ] <upper-bound>
        <assumed-shape-spec> = [ <lower-bound> ] :
        <deferred-shape-spec> = :
        <assumed-size-spec> = [ <explicit-shape-spec-list> , ] [ <lower-bound> : ] *
        <bound> = <specification-expr>

        <int-literal-constant> = <digit-string> [ _ <kind-param> ]
        <digit-string> = <digit> [ <digit> ]..
        <kind-param> = <digit-string> | <scalar-int-constant-name>

    """

    _repr_attr_names = [
        "selector",
        "attrspec",
        "entity_decls",
    ] + Statement._repr_attr_names

    def process_item(self):
        item = self.item
        apply_map = item.apply_map
        clsname = self.__class__.__name__.lower()
        line = item.get_line()
        from .block_statements import Function

        if not line.lower().startswith(clsname):
            i = 0
            j = 0
            for c in line:
                i += 1
                if c == " ":
                    continue
                j += 1
                if j == len(clsname):
                    break
            line = line[:i].replace(" ", "") + line[i:]

        assert line.lower().startswith(clsname), repr((line, clsname))
        line = line[len(clsname) :].lstrip()

        if line.startswith("("):
            i = line.find(")")
            selector = apply_map(line[: i + 1].strip())
            line = line[i + 1 :].lstrip()
        elif line.startswith("*"):
            selector = "*"
            line = line[1:].lstrip()
            if line.startswith("("):
                i = line.find(")")
                selector += apply_map(line[: i + 1].rstrip())
                line = line[i + 1 :].lstrip()
            else:
                m = re.match(r"\d+(_\w+|)|[*]", line)
                if not m:
                    self.isvalid = False
                    return
                i = m.end()
                selector += line[:i].rstrip()
                line = line[i:].lstrip()
        else:
            selector = ""

        fm = Function.match(line)
        if fm:
            l2 = line[: fm.end()]
            m2 = re.match(r".*?\b(?P<name>\w+)\Z", l2)
            if not m2:
                self.isvalid = False
                return
            fname = m2.group("name")
            fitem = item.copy(clsname + selector + " :: " + fname, apply_map=True)
            self.parent.put_item(fitem)
            item.clone(line)
            self.isvalid = False
            return

        if line.startswith(","):
            line = line[1:].lstrip()

        self.raw_selector = selector
        if isinstance(self, Character):
            self.selector = self._parse_char_selector(selector)
        else:
            self.selector = self._parse_kind_selector(selector)

        i = line.find("::")
        if i == -1:
            self.attrspec = []
            self.entity_decls = split_comma(line, self.item)
        else:
            self.attrspec = split_comma(line[:i].rstrip(), self.item)
            self.entity_decls = split_comma(line[i + 2 :].lstrip(), self.item)
        for entity in self.entity_decls:
            if not is_entity_decl(entity):
                self.isvalid = False
                return

        if isinstance(self.parent, Function) and self.parent.name in self.entity_decls:
            assert self.parent.typedecl is None, repr(self.parent.typedecl)
            self.parent.typedecl = self
            self.ignore = True
        if isinstance(self, Type):
            self.name = self.selector[1].lower()
            assert is_name(self.name), repr(self.name)
        else:
            self.name = clsname
        return

    def _parse_kind_selector(self, selector):
        if not selector:
            return "", ""
        length, kind = "", ""
        if selector.startswith("*"):
            length = selector[1:].lstrip()
        else:
            assert selector[0] + selector[-1] == "()", repr(selector)
            l = selector[1:-1].strip()
            if l.lower().startswith("kind"):
                l = l[4:].lstrip()
                if l[0] + l[-1] == "()":
                    kind = "kind" + l
                else:
                    assert l.startswith("="), repr(l)
                    kind = l[1:].lstrip()
            else:
                kind = l
        return length, kind

    def _split_char_selector(self, line):
        """line=``[key=]value`` -> key, value.
        If line does not have name part then return None, value.
        """
        for name in ["len", "kind"]:
            if line[: len(name)].lower() == name:
                value_part = line[len(name) :].lstrip()
                if value_part.startswith("="):
                    return name, value_part[1:].lstrip()
        return None, line

    def _parse_char_selector(self, selector):
        if not selector:
            return "", ""
        if selector.startswith("*"):
            l = selector[1:].lstrip()
            if l.startswith("("):
                if l.endswith(","):
                    l = l[:-1].rstrip()
                assert l.endswith(")"), repr(l)
                l = l[1:-1].strip()
                if l.lower().startswith("len"):
                    l = l[3:].lstrip()[1:].lstrip()
            kind = ""
        else:
            assert selector[0] + selector[-1] == "()", repr(selector)
            l = split_comma(selector[1:-1].strip(), self.item)
            if len(l) == 1:
                l = l[0]
                key, value = self._split_char_selector(l)
                if key == "len":
                    kind, l = "", value
                elif key == "kind":
                    kind, l = value, ""
                else:
                    kind = ""
            else:
                assert len(l) == 2, repr(l)
                key0, value0 = self._split_char_selector(l[0])
                key1, value1 = self._split_char_selector(l[1])
                if key0 == "len":
                    assert key1 in [None, "kind"], repr(key1)
                    l, kind = value0, value1
                elif key0 == "kind":
                    assert key1 == "len", repr(key1)
                    l, kind = value1, value0
                else:
                    assert key0 is None, repr(key0)
                    assert key1 in [None, "kind"], repr(key1)
                    l, kind = value0, value1
        return l, kind

    def tostr(self):
        """Create a text representation of this object and return it"""
        clsname = self.__class__.__name__.upper()
        text = ""
        length, kind = self.selector
        if isinstance(self, Character):
            if length and kind:
                text += "(LEN=%s, KIND=%s)" % (length, kind)
            elif length:
                text += "(LEN=%s)" % (length)
            elif kind:
                text += "(KIND=%s)" % (kind)
        elif isinstance(self, Type):
            text += "(%s)" % (kind)
        elif isinstance(self, Class):
            if kind:
                # For a class declaration, 'kind' is actually the class
                # that the variable is an instance of. Therefore there
                # is no "(KIND=xxx)", just (xxx).
                text += "({0})".format(kind)
        else:
            if length:
                text += "*%s" % (length)
            if kind:
                text += "(KIND=%s)" % (kind)

        return clsname + text

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        s = self.tostr()
        if self.attrspec:
            s += ", " + ", ".join(self.attrspec)
        # If we were to change fparser so that it always produces the
        # '::' separator then we'd simply comment-out the if below.
        if self.attrspec or "=" in str(self.entity_decls):
            s += " ::"
        if self.entity_decls:
            s += " " + ", ".join(self.entity_decls)
        return tab + s

    def __str__(self):
        return self.tofortran()

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return self.selector == other.selector

    def astypedecl(self):
        if self.entity_decls or self.attrspec:
            return self.__class__(self.parent, self.item.copy(self.tostr()))
        return self

    def analyze(self):
        if not self.entity_decls:
            return
        variables = self.parent.a.variables
        typedecl = self.astypedecl()
        attrspec = self.attrspec[:]
        access_spec_lst = [a for a in attrspec if a.lower() in ["private", "public"]]
        if access_spec_lst:
            access_spec = access_spec_lst[0]
            attrspec.remove(access_spec)
        else:
            access_spec = None
        for item in self.entity_decls:
            name, array_spec, char_length, value = self._parse_entity(item)
            var = self.parent.get_variable(name)
            var.add_parent(self)
            if char_length:
                var.set_length(char_length)
            var.set_type(typedecl)
            var.update(self.attrspec)
            if array_spec:
                var.set_bounds(array_spec)
            if value:
                var.set_init(value)
            if access_spec is not None:
                l = getattr(self.parent.a, access_spec.lower() + "_id_list")
                l.append(name)
            var.analyze()
        return

    def _parse_entity(self, line):
        m = name_re(line)
        assert m, repr((line, self.item, self.__class__.__name__))
        name = line[: m.end()]
        line = line[m.end() :].lstrip()
        array_spec = None
        char_length = None
        value = None
        if line:
            item = self.item.copy(line)
            line = item.get_line()
            if line.startswith("("):
                i = line.find(")")
                assert i != -1, repr(line)
                array_spec = parse_array_spec(line[1:i].strip(), item)
                line = line[i + 1 :].lstrip()

            if line.startswith("*"):
                i = line.find("=")
                if i == -1:
                    char_length = item.apply_map(line[1:].lstrip())
                    line = ""
                else:
                    char_length = item.apply_map(line[1:i].strip())
                    line = line[i:]
            if line.startswith("="):
                value = item.apply_map(line[1:].lstrip())
        return name, array_spec, char_length, value

    def get_zero_value(self):
        raise NotImplementedError(repr(self.__class__.__name__))

    def assign_expression(self, name, value):
        return "%s = %s" % (name, value)

    def get_kind(self):
        return self.selector[1] or self.default_kind

    def get_length(self):
        return self.selector[0] or 1

    def get_byte_size(self):
        length, kind = self.selector
        if length:
            return int(length)
        if kind:
            return int(kind)
        return self.default_kind

    def is_intrinsic(self):
        return not isinstance(self, (Type, Class))

    def is_derived(self):
        return isinstance(self, Type)

    def is_numeric(self):
        return isinstance(
            self, (Integer, Real, DoublePrecision, Complex, DoubleComplex, Byte)
        )

    def is_nonnumeric(self):
        return isinstance(self, (Character, Logical))


class Integer(TypeDeclarationStatement):
    match = re.compile(r"integer\b", re.I).match
    default_kind = 4

    def get_zero_value(self):
        kind = self.get_kind()
        if kind == self.default_kind:
            return "0"
        return "0_%s" % (kind)


class Real(TypeDeclarationStatement):
    match = re.compile(r"real\b", re.I).match
    default_kind = 4

    def get_zero_value(self):
        kind = self.get_kind()
        if kind == self.default_kind:
            return "0.0"
        return "0_%s" % (kind)


class DoublePrecision(TypeDeclarationStatement):
    match = re.compile(r"double\s*precision\b", re.I).match
    default_kind = 8

    def get_byte_size(self):
        return self.default_kind

    def get_zero_value(self):
        return "0.0D0"


class Complex(TypeDeclarationStatement):
    match = re.compile(r"complex\b", re.I).match
    default_kind = 4

    def get_byte_size(self):
        length, kind = self.selector
        if length:
            return int(length)
        if kind:
            return 2 * int(kind)
        return 2 * self.default_kind

    def get_zero_value(self):
        kind = self.get_kind()
        if kind == self.default_kind:
            return "(0.0, 0.0)"
        return "(0.0_%s, 0.0_%s)" % (kind, kind)

    def get_part_typedecl(self):
        bz = self.get_byte_size() / 2
        return Real(self.parent, self.item.copy("REAL*%s" % (bz)))


class DoubleComplex(TypeDeclarationStatement):
    # not in standard
    match = re.compile(r"double\s*complex\b", re.I).match
    default_kind = 8

    def get_byte_size(self):
        return 2 * self.default_kind

    def get_zero_value(self):
        return "(0.0D0,0.0D0)"


class Logical(TypeDeclarationStatement):
    match = re.compile(r"logical\b", re.I).match
    default_kind = 4

    def get_zero_value(self):
        return ".FALSE."


class Character(TypeDeclarationStatement):
    match = re.compile(r"character\b", re.I).match
    default_kind = 1

    def get_zero_value(self):
        return "''"


class Byte(TypeDeclarationStatement):
    # not in standard
    match = re.compile(r"byte\b", re.I).match
    default_kind = 1

    def get_zero_value(self):
        return "0"


class Type(TypeDeclarationStatement):
    match = re.compile(r"type\s*\(", re.I).match

    def get_zero_value(self):
        type_decl = self.get_type_decl(self.name)
        component_names = type_decl.a.component_names
        components = type_decl.a.components
        l = []
        for name in component_names:
            var = components[name]
            l.append(var.typedecl.get_zero_value())
        return "%s(%s)" % (type_decl.name, ", ".join(l))

    def get_kind(self):
        # See 4.5.2, page 48
        raise NotImplementedError(repr(self.__class__.__name__))


TypeStmt = Type


class Class(TypeDeclarationStatement):
    match = re.compile(r"class\s*\(", re.I).match


class Implicit(Statement):
    """
    Class capturing various forms of IMPLICIT statement.

    ::

        IMPLICIT <implicit-spec-list>
        IMPLICIT NONE
        <implicit-spec> = <declaration-type-spec> ( <letter-spec-list> )
        <letter-spec> = <letter> [ - <letter> ]

    """

    match = re.compile(r"implicit\b", re.I).match

    letters = string.ascii_lowercase

    def process_item(self):
        line = self.item.get_line()[8:].lstrip()
        if line.lower() == "none":
            self.items = []
            return
        items = []
        for item in split_comma(line, self.item):
            i = item.find("(")
            assert i != -1 and item.endswith(")"), repr(item)
            specs = []
            for spec in split_comma(item[i + 1 : -1].strip(), self.item):
                if "-" in spec:
                    s, e = spec.lower().split("-")
                    s = s.strip()
                    e = e.strip()
                    assert s in self.letters and e in self.letters, repr((s, e))
                else:
                    e = s = spec.lower().strip()
                    assert s in self.letters, repr((s, e))
                specs.append((s, e))
            tspec = item[:i].rstrip()
            stmt = None
            for cls in declaration_type_spec:
                if cls.match(tspec):
                    stmt = cls(self, self.item.copy(tspec))
                    if stmt.isvalid:
                        break
            assert stmt is not None, repr((item, line))
            items.append((stmt, specs))
        self.items = items
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        if not self.items:
            return tab + "IMPLICIT NONE"
        l = []
        for stmt, specs in self.items:
            l1 = []
            for s, e in specs:
                if s == e:
                    l1.append(s)
                else:
                    l1.append(s + "-" + e)
            l.append("%s ( %s )" % (stmt.tostr(), ", ".join(l1)))
        return tab + "IMPLICIT " + ", ".join(l)

    def analyze(self):
        """
        Analyze the Implicit statments constructed by the parser and
        set-up the associated implicit_rules belonging to the parent
        of this object in the AST.
        """
        implicit_rules = self.parent.a.implicit_rules
        if not self.items:
            if implicit_rules:
                self.warning(
                    "overriding previously set implicit rule mapping"
                    " %r." % (implicit_rules)
                )
            self.parent.a.implicit_rules = None
            return
        if implicit_rules is None:
            self.warning("overriding previously set IMPLICIT NONE")
            self.parent.a.implicit_rules = implicit_rules = {}
        for stmt, specs in self.items:
            for start, end in specs:
                start_idx = string.ascii_lowercase.index(start.lower())
                end_idx = string.ascii_lowercase.index(end.lower())
                for lchar in string.ascii_lowercase[start_idx : end_idx + 1]:
                    implicit_rules[lchar] = stmt
        return


intrinsic_type_spec = [
    Integer,
    Real,
    DoublePrecision,
    Complex,
    DoubleComplex,
    Character,
    Logical,
    Byte,
]
declaration_type_spec = intrinsic_type_spec + [TypeStmt, Class]



# Copyright (c) 2023 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R830 : This file tests the support for the
loop-control rule.

"""

import pytest
from fparser.two.Fortran2003 import Loop_Control
from fparser.two.utils import NoMatchError


def test_start_end_space_while():
    """Test that there is a match if the string contains white space at
    the start and end and that the tostr() output is as expected. Also
    test matching to the while form of this rule.

    """
    result = Loop_Control(" while (.true.) ")
    assert isinstance(result, Loop_Control)
    assert str(result) == "WHILE (.TRUE.)"


def test_delim():
    """Test that there is a match if the string contains an optional
    delimiter at the start and that the tostr() output is as expected.

    """
    result = Loop_Control(" , while (.true.) ")
    assert isinstance(result, Loop_Control)
    assert str(result) == ", WHILE (.TRUE.)"


def test_repmap():
    """Test matching when the while logical expresssion contains
    brackets. This tests the use of the string_replace_map() function.

    """
    result = Loop_Control(" , while (((a .or. b) .and. (c .or. d))) ")
    assert isinstance(result, Loop_Control)
    assert str(result) == ", WHILE (((a .OR. b) .AND. (c .OR. d)))"


def test_counter():
    """Test matching to the counter form of this rule and that the tostr()
    output is as expected.

    """
    # Lower-bound and upper-bound only
    result = Loop_Control("idx = start,stop")
    assert isinstance(result, Loop_Control)
    assert str(result) == "idx = start, stop"
    # Lower-bound, upper-bound and step
    result = Loop_Control("idx = start,stop,step")
    assert isinstance(result, Loop_Control)
    assert str(result) == "idx = start, stop, step"
    # Bounds are integer expressions
    result = Loop_Control("idx = ((s+2)-q),(p*m)/4,a+b+c")
    assert isinstance(result, Loop_Control)
    assert str(result) == "idx = ((s + 2) - q), (p * m) / 4, a + b + c"


@pytest.mark.parametrize(
    "string",
    [
        "",
        " ",
        ": while(.true.)",
        "whil (.true.)",
        "while .true.",
        "while ()",
        "while( )",
        "while (.true ",
        "while ())",
        "while('text')",
        " == ",
        " = ",
        "idx=",
        "=1,2",
        "idx=1",
        "idx=1,2,3,4",
        "1=1,2",
        "idx=1,.false.",
    ],
)
def test_invalid(string):
    """Test that there is no match for various invalid input strings."""
    with pytest.raises(NoMatchError):
        _ = Loop_Control(string)



# Copyright (c) 2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
pytest tests for Fortran2003 rule R501 - Type Declaration Statement.

TODO #318 - these tests need extending to fully cover the
Fortran2003.Type_Declaration_Stmtclass. They will also need to be
extended as part of #259 as that will add support for the various
constraints that apply to R501.

"""

import pytest
from fparser.two import Fortran2003
from fparser.two.symbol_table import SYMBOL_TABLES


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("table_name", ["", "test_mod"])
def test_type_declaration_stmt(table_name):
    """Various tests for the type declaration statement (R501). We test both
    with and without an existing scoping region."""
    if table_name:
        SYMBOL_TABLES.enter_scope(table_name)
    table = SYMBOL_TABLES.current_scope

    tcls = Fortran2003.Type_Declaration_Stmt
    obj = tcls("integer a")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "INTEGER :: a"
    assert (
        repr(obj) == "Type_Declaration_Stmt(Intrinsic_Type_Spec('INTEGER', "
        "None), None, Entity_Decl_List(',', (Entity_Decl(Name('a'), None, "
        "None, None),)))"
    )
    if table:
        assert "a" in table._data_symbols

    obj = tcls("integer ,dimension(2):: b*3")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "INTEGER, DIMENSION(2) :: b*3"
    if table:
        assert "b" in table._data_symbols

    obj = tcls("real c")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "REAL :: c"
    assert (
        repr(obj) == "Type_Declaration_Stmt(Intrinsic_Type_Spec('REAL', None), None, "
        "Entity_Decl_List(',', (Entity_Decl(Name('c'), None, None, "
        "None),)))"
    )
    if table:
        assert "c" in table._data_symbols

    obj = tcls("REAL D( LDA, * ), E( LDB, * )")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "REAL :: D(LDA, *), E(LDB, *)"
    if table:
        assert "d" in table._data_symbols
        assert "e" in table._data_symbols

    obj = tcls("DOUBLE PRECISION   ALPHA, BETA")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "DOUBLE PRECISION :: ALPHA, BETA"
    if table:
        assert "alpha" in table._data_symbols
        assert "beta" in table._data_symbols

    obj = tcls("logical,parameter:: T=.true.")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "LOGICAL, PARAMETER :: T = .TRUE."
    if table:
        assert "t" in table._data_symbols

    obj = tcls("character(n),private:: x(n)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CHARACTER(LEN = n), PRIVATE :: x(n)"
    if table:
        assert "x" in table._data_symbols

    obj = tcls("character(lenmax),private:: y(n)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CHARACTER(LEN = lenmax), PRIVATE :: y(n)"
    if table:
        assert "y" in table._data_symbols



# Copyright (c) 2022 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R856

error-stop-stmt is ERROR STOP [ stop-code ]
"""

import pytest
from fparser.two.utils import NoMatchError, walk
from fparser.two.Fortran2008 import Error_Stop_Stmt
from fparser.api import get_reader


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("string", ["ERROR STOP", "error stop"])
def test_simple(string):
    """Test that error-stop-stmt without stop-code matches."""
    result = Error_Stop_Stmt(string)
    assert str(result) == "ERROR STOP"


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("stop_code", ["1", "A", "err_code"])
def test_stop_code(stop_code):
    """Test that error-stop-stmt with stop-code matches."""
    result = Error_Stop_Stmt("ERROR STOP {}".format(stop_code))
    assert str(result) == "ERROR STOP {}".format(stop_code)


@pytest.mark.usefixtures("f2008_create")
def test_error1():
    """Test error-stop-stmt is not matched for wrong syntax."""
    with pytest.raises(NoMatchError) as excinfo:
        _ = Error_Stop_Stmt("ERROR () STOP")
    assert "Error_Stop_Stmt: 'ERROR () STOP'" in str(excinfo.value)


def test_functional1(f2008_parser):
    """Test error-stop-stmt is matched in a subroutine."""
    tree = f2008_parser(get_reader("""\
subroutine my_abort
error stop
end subroutine my_abort
    """))
    assert walk(tree, Error_Stop_Stmt)
    assert "ERROR STOP" in str(tree)


def test_functional2(f2008_parser):
    """Test error-stop-stmt is matched in a if-stmt."""
    tree = f2008_parser(get_reader("""\
subroutine my_abort(err_code)
implicit none
integer, intent(in) :: err_code
if (err_code /= 0) error stop err_code
end subroutine my_abort
    """))
    assert walk(tree, Error_Stop_Stmt)
    assert "ERROR STOP err_code" in str(tree)



# Copyright (c) 2021 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R724 : This file tests the support for a
Fortran logical expression.

"""

import pytest
from fparser.two.Fortran2003 import (
    Logical_Expr,
    Logical_Literal_Constant,
    Equiv_Operand,
    Or_Operand,
)
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2003_create")
def test_simple_case():
    """Test that a simple expression gives the expected result."""
    result = Logical_Expr(".true.")
    assert isinstance(result, Logical_Literal_Constant)
    assert str(result) == ".TRUE."
    assert repr(result) == "Logical_Literal_Constant('.TRUE.', None)"


@pytest.mark.usefixtures("f2003_create", "fake_symbol_table")
def test_complicated_case():
    """Test that a more complicated expression gives the expected
    result.

    """
    result = Logical_Expr("(f0 .lt. f1) .and. abs(x1-x0) .gt. abs(x2) .or.  .not. root")
    assert isinstance(result, Equiv_Operand)
    assert str(result) == (
        "(f0 .LT. f1) .AND. ABS(x1 - x0) .GT. ABS(x2) " ".OR. .NOT. root"
    )
    assert repr(result).replace("u'", "'") == (
        "Equiv_Operand(Or_Operand(Parenthesis('(', Level_4_Expr(Name('f0'), "
        "'.LT.', Name('f1')), ')'), '.AND.', Level_4_Expr(Intrinsic_Function"
        "_Reference(Intrinsic_Name('ABS'), Actual_Arg_Spec_List(',', (Level_"
        "2_Expr(Name('x1'), '-', Name('x0')),))), '.GT.', Intrinsic_Function"
        "_Reference(Intrinsic_Name('ABS'), Actual_Arg_Spec_List(',', (Name('"
        "x2'),))))), '.OR.', And_Operand('.NOT.', Name('root')))"
    )


@pytest.mark.usefixtures("f2003_create", "fake_symbol_table")
def test_string_comparison_with_backslash():
    """
    Check that a logical expression involving comparison with a string containing
    a backslash is parsed correctly.

    """
    result = Logical_Expr("MetFolder(L:L) == '\\' .and. L <= MaxFileNameLength")
    assert isinstance(result, Or_Operand)
    assert str(result) == "MetFolder(L : L) == '\\' .AND. L <= MaxFileNameLength"


@pytest.mark.parametrize(
    "string", ["1", "b'1010'", "o'7070'", "h'f0f0'", "1.0", "(1.0,1.0)", "'hello'"]
)
@pytest.mark.usefixtures("f2003_create")
def test_c705(string):
    """Check that invalid literal constants do not match. Note, there are
    many other cases that are not currently checked in fparser.

    In theory -1 should become a signed_int_literal_constant and then
    fail to match. Similarly -1.0 should become a
    Signed_Real_Literal_Constant and then fail to match. However the
    "-" is treated as a unary expression so this never happens.

    """
    with pytest.raises(NoMatchError):
        _ = Logical_Expr(string)



# This Python file uses the following encoding: utf-8
# Copyright (c) 2020 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""pytest module for the Fortran2003 Write Statement - R911."""

import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2003 import Write_Stmt, Io_Control_Spec_List


@pytest.mark.usefixtures("f2003_create")
def test_write_stmt():
    """Tests for various forms of WRITE statement (R911)."""
    tcls = Write_Stmt
    obj = tcls('write (123)"hey"')
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'WRITE(123) "hey"'
    assert repr(obj).replace("u'", "'") == (
        "Write_Stmt(Io_Control_Spec_List(',', (Io_Control_Spec(None, "
        "Int_Literal_Constant('123', None)),)), Output_Item_List(',', "
        "(Char_Literal_Constant('\"hey\"', None),)))"
    )

    obj = tcls('WRITE (*,"(I3)") my_int')
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == 'WRITE(*, "(I3)") my_int'
    assert repr(obj).replace("u'", "'") == (
        "Write_Stmt(Io_Control_Spec_List(',', (Io_Control_Spec(None, "
        "Io_Unit('*')), Io_Control_Spec(None, "
        "Char_Literal_Constant('\"(I3)\"', None)))), Output_Item_List(',', "
        "(Name('my_int'),)))"
    )

    obj = tcls("WRITE (*,namtest)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "WRITE(*, namtest)"
    assert repr(obj).replace("u'", "'") == (
        "Write_Stmt(Io_Control_Spec_List(',', "
        "(Io_Control_Spec(None, Io_Unit('*')), Io_Control_Spec(None, "
        "Name('namtest')))), None)"
    )

    # Test when format specifier contains an '=' character
    iolist = Io_Control_Spec_List("*,'(5X,\"q_mesh =\",4F12.8)'")
    assert isinstance(iolist, Io_Control_Spec_List)
    obj = tcls("WRITE(*,'(5X,\"q_mesh =\",1F12.8)') 1.d0")
    assert isinstance(obj, tcls)
    assert repr(obj).replace("u'", "'") == (
        "Write_Stmt(Io_Control_Spec_List(',', (Io_Control_Spec(None, "
        "Io_Unit('*')), Io_Control_Spec(None, "
        "Char_Literal_Constant('\\'(5X,\"q_mesh =\",1F12.8)\\'', None)))), "
        "Output_Item_List(',', (Real_Literal_Constant('1.D0', None),)))"
    )

    obj = tcls("WRITE(*,FMT='(5X,\"q_mesh =\",1F12.8)') 1.d0")
    assert isinstance(obj, tcls)
    assert repr(obj).replace("u'", "'") == (
        "Write_Stmt(Io_Control_Spec_List(',', (Io_Control_Spec(None, "
        "Io_Unit('*')), Io_Control_Spec('FMT', "
        "Char_Literal_Constant('\\'(5X,\"q_mesh =\",1F12.8)\\'', None)))), "
        "Output_Item_List(',', (Real_Literal_Constant('1.D0', None),)))"
    )

    # Format specifier contains an '=' and is built using concatenation
    obj = tcls('''WRITE (6, '("write = some=""'//'text'//'""")')''')
    assert isinstance(obj, tcls)
    assert str(obj) == '''WRITE(6, '("write = some=""' // 'text' // '""")')'''
    obj_repr = repr(obj)
    obj_repr = obj_repr.replace('u"', '"')
    assert obj_repr.replace("u'", "'") == (
        "Write_Stmt(Io_Control_Spec_List(',', "
        "(Io_Control_Spec(None, Int_Literal_Constant('6', None)), "
        "Io_Control_Spec(None, Level_3_Expr(Level_3_Expr("
        "Char_Literal_Constant('\\'(\"write = some=\"\"\\'', None), '//', "
        "Char_Literal_Constant(\"'text'\", None)), '//', "
        "Char_Literal_Constant('\\'\"\"\")\\'', None))))), None)"
    )


@pytest.mark.usefixtures("f2003_create")
def test_named_unit_before_fmt_error():
    """Check that we reject a WRITE that names the io-unit argument but
    still has a positional format argument (containing an '=').
    TODO #267. This test needs expanding and probably moving to a file
    dedicated to R913 and its (many) constraints.

    """
    tcls = Write_Stmt
    # Cannot have an un-named (positional) argument after a named argument
    with pytest.raises(NoMatchError):
        tcls('''WRITE (UNIT=6, '("write some=""'//'text'//'""")')''')



# Copyright (c) 2020 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R465: this file tests that both (/.../) and
[...] are supported when specifying an array constructor.

"""

import pytest
from fparser.common.readfortran import FortranStringReader
from fparser.two import Fortran2003


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("left, right", [("(/", "/)"), ("[", "]")])
def test_brackets_array_constructor(left, right):
    """Test parsing of array constructor specified with both valid types of
    bracket."""
    fcode = "array = {0} 1, 2, 3{1}".format(left, right)
    reader = FortranStringReader(fcode)
    ast = Fortran2003.Assignment_Stmt(reader)
    assert isinstance(ast, Fortran2003.Assignment_Stmt)
    assert isinstance(ast.children[2], Fortran2003.Array_Constructor)
    assert isinstance(ast.children[2].children[1], Fortran2003.Ac_Value_List)
    assert "array = {0}1, 2, 3{1}".format(left, right) in str(ast)
    # Invalid content between valid brackets
    fcode = "array = {0}call hello(){1}".format(left, right)
    reader = FortranStringReader(fcode)
    ast = Fortran2003.Assignment_Stmt(reader)
    assert ast is None


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("example", ["[1,2/)", "[1,2[]", "(/1,3]", "(/1,3)"])
def test_array_constructor_no_match(example):
    """Check that incorrect constructors are not matched."""
    fcode = "array = " + example
    reader = FortranStringReader(fcode)
    ast = Fortran2003.Assignment_Stmt(reader)
    assert ast is None



# Modified work Copyright (c) 2017-2022 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Associate_Construct
from fparser.two.utils import FortranSyntaxError


@pytest.mark.parametrize(
    "code, expected_string",
    [
        (
            """\
            ASSOCIATE ( Z => EXP(-(X**2+Y**2)) * COS(THETA) )
                PRINT *, A+Z, A-Z
            END ASSOCIATE
            """,
            "ASSOCIATE(Z => EXP(- (X ** 2 + Y ** 2)) * COS(THETA))\n"
            "  PRINT *, A + Z, A - Z\nEND ASSOCIATE",
        ),
        (
            """\
            name:ASSOCIATE ( XC => AX%B(I,J)%C )
                XC%DV = XC%DV + PRODUCT(XC%EV(1:N))
            END ASSOCIATE name
            """,
            "name:ASSOCIATE(XC => AX % B(I, J) % C)\n  XC % DV = XC % DV + "
            "PRODUCT(XC % EV(1 : N))\nEND ASSOCIATE name",
        ),
        (
            """\
            ASSOCIATE ( W => RESULT(I,J)%W, ZX => AX%B(I,J)%D, ZY => AY%B(I,J)%D )
                W = ZX*X + ZY*Y
            END ASSOCIATE
            """,
            "ASSOCIATE(W => RESULT(I, J) % W, ZX => AX % B(I, J) % D, ZY => "
            "AY % B(I, J) % D)\n  W = ZX * X + ZY * Y\nEND ASSOCIATE",
        ),
    ],
)
def test_associate_construct(fake_symbol_table, code, expected_string):
    """Test some basic 'associate' constructs are parsed correctly"""
    obj = Associate_Construct(get_reader(code))
    assert isinstance(obj, Associate_Construct), repr(obj)
    assert str(obj) == expected_string


def test_end_block_missing_name(f2003_create, fake_symbol_table):
    """Check that a named associate block has a name at the end"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Associate_Construct(get_reader("""\
                name:associate (xc => ax%b(i,j)%c)
                    xc%dv = xc%dv + product(xc%ev(1:n))
                end associate
                """))
    assert exc_info.value.args[0].endswith("Expecting name 'name' but none given")


def test_end_block_wrong_name(f2003_create, fake_symbol_table):
    """Check that a named associate block has the correct name at the end"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Associate_Construct(get_reader("""\
                name:associate (xc => ax%b(i,j)%c)
                    xc%dv = xc%dv + product(xc%ev(1:n))
                end associate wrong
                """))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023-2024, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""Test Fortran 2008 rule R1003

format-item is [ r ] data-edit-desc
            or control-edit-desc
            or char-string-edit-desc
            or [ r ] ( format-item-list )
            or format-item-c1002
            or hollerith-item
            or * ( format-item-list )

Extends the Fortran 2003 rule R1003 with the additional unlimited
format repeat specifier ``*`` before a parenthesised format-item-list.

"""

import pytest

from fparser.common.readfortran import FortranStringReader
from fparser.two.Fortran2008 import Format_Item
from fparser.two.utils import NoMatchError


def test_unlimited_repeat_basic(f2008_create):
    """Test basic unlimited format repeat: *(I5)."""
    obj = Format_Item("*(I5)")
    assert isinstance(obj, Format_Item)
    assert str(obj) == "*(I5)"


def test_unlimited_repeat_multiple_items(f2008_create):
    """Test unlimited format repeat with multiple items: *(1X,A12)."""
    obj = Format_Item("*(1X,A12)")
    assert isinstance(obj, Format_Item)
    assert str(obj) == "*(1X, A12)"


def test_unlimited_repeat_pe_descriptor(f2008_create):
    """Test unlimited format repeat with PE descriptor: *(1X,1PE12.5)."""
    obj = Format_Item("*(1X,1PE12.5)")
    assert isinstance(obj, Format_Item)
    assert str(obj) == "*(1X, 1P, E12.5)"


def test_f2003_match(f2008_create):
    """Test that F2003 format items still match via the F2008 class."""
    obj = Format_Item("3(I5)")
    assert isinstance(obj, Format_Item)
    assert str(obj) == "3(I5)"


def test_format_stmt_no_match():
    """Test various cases that fail to match."""
    with pytest.raises(NoMatchError):
        _ = Format_Item("")
    with pytest.raises(NoMatchError):
        _ = Format_Item("  ")
    with pytest.raises(NoMatchError):
        _ = Format_Item("*(")
    with pytest.raises(NoMatchError):
        _ = Format_Item("*(1X")


def test_format_stmt_unlimited_repeat(f2008_parser):
    """Test unlimited format repeat in a full FORMAT statement."""

    reader = FortranStringReader("subroutine t()\n  1 format(*(I5))\nend subroutine")
    tree = f2008_parser(reader)
    assert tree is not None


def test_format_stmt_mixed_items(f2008_parser):
    """Test FORMAT with regular items followed by unlimited repeat."""

    reader = FortranStringReader(
        "subroutine t()\n  1 format('!',A12,*(1X,A12))\nend subroutine"
    )
    tree = f2008_parser(reader)
    assert tree is not None


def test_format_stmt_pe_descriptor(f2008_parser):
    """Test FORMAT with unlimited repeat and PE edit descriptor."""

    reader = FortranStringReader(
        "subroutine t()\n  2 format(*(1X,1PE12.5))\nend subroutine"
    )
    tree = f2008_parser(reader)
    assert tree is not None



# Copyright (c) 2022-2023 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing pytest tests for the support of the Fortran2008
Critical construct."""

import pytest

from fparser.api import get_reader
from fparser.two.Fortran2003 import Assignment_Stmt
from fparser.two.Fortran2008 import Critical_Construct, Critical_Stmt, End_Critical_Stmt
from fparser.two.utils import FortranSyntaxError


def test_critical():
    """Test that a basic critical construct is correctly constructed."""
    critical = Critical_Construct(get_reader("""\
            critical
               a = 1 + b
            end critical
            """))
    assert isinstance(critical.children[0], Critical_Stmt)
    assert isinstance(critical.children[1], Assignment_Stmt)
    assert isinstance(critical.children[2], End_Critical_Stmt)
    assert critical.children[0].get_start_name() is None
    assert "CRITICAL\n  a = 1 + b\nEND CRITICAL" in str(critical)


def test_named_critical():
    """Test that a named critical construct is matched correctly and that
    its name can be queried."""
    critical = Critical_Construct(get_reader("""\
            foo: critical
               a = 1 + b
            end critical foo
            """))
    assert critical.children[0].get_start_name() == "foo"
    assert "foo:CRITICAL\n  a = 1 + b\nEND CRITICAL foo" in str(critical)


def test_end_critical_missing_start_name():  # C809
    """Check that a critical construct with an end name but no start name
    results in a syntax error (C809)."""
    with pytest.raises(FortranSyntaxError) as err:
        Critical_Construct(get_reader("""\
                critical
                end critical foo
                """))
    assert "Name 'foo' has no corresponding starting name" in str(err)


def test_end_critical_missing_end_name():  # C809
    """Test that a named critical construct with the name omitted from
    the end critical results in a syntax error (C809)."""
    with pytest.raises(FortranSyntaxError) as err:
        Critical_Construct(get_reader("""\
                foo: critical
                end critical
                """))
    assert "Expecting name 'foo' but none given" in str(err)


def test_end_critical_wrong_name():  # C809
    """Test that mismatched start and end names result in a syntax error (C809)"""
    with pytest.raises(FortranSyntaxError) as err:
        Critical_Construct(get_reader("""\
                foo: critical
                end critical bar
                """))
    assert "Expecting name 'foo', got 'bar'" in str(err)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2024-2025, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ------------------------------------------------------------------------------
# Author: Joerg Henrichs, Bureau of Meteorology

"""This file contains a script that will remove any private or protected
declaration in a Fortran file. This is used by PSyclone's kernel extraction
and driver creation feature, see
https://psyclone.readthedocs.io/en/latest/psyke.html.
for details.

Usage:  make_public.py  file1.f90
"""

import sys

from fparser.common.readfortran import FortranFileReader
from fparser.two.Fortran2003 import (
    Access_Stmt,
    Access_Spec,
    Attr_Spec,
    Binding_Private_Stmt,
    Private_Components_Stmt,
    Protected_Stmt,
)
from fparser.two.parser import ParserFactory
from fparser.two.Fortran2003 import Type_Declaration_Stmt
from fparser.two.utils import walk


# -----------------------------------------------------------------------------
def remove_private(filename):
    """Simple function that removes all private and protected declarations.
    :param str filename: the file in which to remove private and protected
    """
    # Do not filter out comments, since this would remove UM-style
    # 'depends on' comments for dependencies that are otherwise not detected.
    reader = FortranFileReader(filename, ignore_comments=False)
    parser = ParserFactory().create(std="f2008")
    parse_tree = parser(reader)
    # A useful print to see the actual rules
    # print(repr(parse_tree))

    # Loop over all access and protected statements. Note that a
    # `protected_stmt` is not an access statement, so it needs to
    # be listed additionally:
    for node in walk(
        parse_tree,
        (Access_Stmt, Protected_Stmt, Private_Components_Stmt, Binding_Private_Stmt),
    ):
        # A Private_Components_Stms has no items:
        if isinstance(
            node, (Binding_Private_Stmt, Private_Components_Stmt)
        ) or node.items[0] in ["PRIVATE", "PROTECTED"]:
            # Find the node in the parent, and remove it:
            node.parent.children.remove(node)

    for node in walk(parse_tree, Access_Spec):
        if str(node) != "PRIVATE":
            continue

        # In some cases it might be required to leave certain declarations
        # private. One example is the variable `ModuleName`` which is used
        # with DrHook to specify the name of the module. If at the same time
        # another modules does a wild-card import, this importing module will
        # have two definitions of ModuleName, its own and the imported one.
        # You can add any variable names that should not be turned public
        # to the `names_to_ignore` list, and the listed names will not be
        # made public.
        names_to_ignore = []
        if isinstance(node.parent.parent, Type_Declaration_Stmt):
            # type-declaration-stmt is declaration-type-spec [
            #     [ , attr-spec ]... :: ] entity-decl-list
            entity_decl = node.parent.parent.items[-1].items[0]
            if str(entity_decl.items[0]).lower() in names_to_ignore:
                # Don't change a variable in the names_to_ignore list
                continue
        node.string = "PUBLIC"

    all_nodes = list(walk(parse_tree, Attr_Spec))
    for node in all_nodes:
        if str(node) == "PROTECTED":
            # This is a tuple, so we can't simply remove the attribute
            node.parent.items = tuple(i for i in node.parent.items if i is not node)
            # If all items in the Attr_Spec are removed, we need to replace the
            # Attr_Spec in the parent-parent (Type_Declaration) with None,
            # otherwise fparser will create e.g. `real, :: a`
            if len(node.parent.items) == 0:
                # Again all tuples, which we can't modify, so we need to
                # recreate the tuple but replace the attr_spec with None
                type_decl = node.parent.parent
                type_decl.items = (type_decl.items[0], None, type_decl.items[2])

    return parse_tree


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    user_filename = sys.argv[1]
    modified_parse_tree = remove_private(user_filename)
    print(modified_parse_tree)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#############################################################################
# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
"""
Tests the fparser.one.parsefortran module.
"""

import pytest
import fparser.one.parsefortran
import fparser.common.readfortran


def test_log_empty(log):
    """
    Tests that a reader without next() method causes an event to be logged.
    """

    class EmptyReader:
        """
        A faux reader with no next() method.
        """

        id = "thingumy"

    unit_under_test = fparser.one.parsefortran.FortranParser(EmptyReader())
    unit_under_test.analyze()
    assert log.messages == {
        "debug": [],
        "info": ["Nothing to analyze."],
        "warning": [],
        "error": [],
        "critical": [],
    }
    unit_under_test.cache.clear()


def test_log_cache(log):
    """
    Tests that using a cached reader object logs an event.
    """

    class Readerlike:
        """
        Dummy reader class, the only purpose of which is to have an id and not
        cause the parser to fail.
        """

        id = "thisun"

        def next(self):
            """
            Simple non-failure-causing method.
            """
            yield "NOT A THING"
            raise StopIteration

    # Expect everything to go okay, no log messages.
    log.reset()
    parser = fparser.one.parsefortran.FortranParser(Readerlike())
    assert log.messages == {
        "debug": [],
        "info": [],
        "warning": [],
        "error": [],
        "critical": [],
    }

    # This time we should use a cached log.
    parser = fparser.one.parsefortran.FortranParser(Readerlike())
    assert log.messages == {
        "debug": [],
        "info": ["using cached thisun"],
        "warning": [],
        "error": [],
        "critical": [],
    }
    parser.cache.clear()


def test_log_failure(log, monkeypatch):
    """
    Tests that an unexpected read failure causes an event to be logged.
    """

    def faulty_next(self, ignore_comments=False):
        """
        Raies any old exception.
        """
        raise Exception("That" "s all folks!")

    monkeypatch.setattr(
        "fparser.common.readfortran.FortranStringReader.next", faulty_next
    )
    reader = fparser.common.readfortran.FortranStringReader("")
    unit_under_test = fparser.one.parsefortran.FortranParser(reader)
    with pytest.raises(Exception):
        unit_under_test.parse()
    assert log.messages["debug"][0].startswith("An error occurred during parsing.")
    assert log.messages["info"] == []
    assert log.messages["warning"] == []
    assert log.messages["error"] == []
    assert log.messages["critical"][0].startswith("While processing")
    assert log.messages["critical"][1] == "STOPPED PARSING"
    unit_under_test.cache.clear()


def test_pyf():
    """
    Tests inherited from implementation code.
    """
    string = """
python module foo
  interface tere
    subroutine bar
    real r
    end subroutine bar
  end interface tere
end python module foo
"""
    expected = [
        "  PYTHONMODULE foo",
        "    INTERFACE tere",
        "      SUBROUTINE bar()",
        "        REAL r",
        "      END SUBROUTINE bar",
        "    END INTERFACE tere",
        "  END PYTHONMODULE foo",
    ]

    reader = fparser.common.readfortran.FortranStringReader(string)
    reader.set_format(fparser.common.sourceinfo.FortranFormat.from_mode("pyf"))
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()
    caught = parser.block.tofortran().splitlines()
    assert caught[0][:13] == "!BEGINSOURCE "
    assert caught[1:] == expected


def test_free90():
    """
    Tests inherited from implementation code.
    """
    string = """
module foo

   subroutine bar
    real r
    if ( pc_get_lun() .ne. 6) &
    write ( pc_get_lun(), '( &
    & /, a, /, " p=", i4, " stopping c_flag=", a, &
    & /, " print unit=", i8)') &
    trim(title), pcpsx_i_pel(), trim(c_flag), pc_get_lun()
    if (.true.) then
      call smth
    end if
    aaa : if (.false.) then
    else if (a) then
    else
    end if aaa
    hey = 1
    end subroutine bar
    abstract interface

    end interface

end module foo
"""
    expected = [
        "  MODULE foo",
        "    SUBROUTINE bar()",
        "      REAL r",
        "      IF (pc_get_lun() .ne. 6)"
        + ' WRITE (pc_get_lun(), \'(  /, A, /, " P = ", i4,'
        + ' " stopping c_flag=", a,  /, " print unit=", i8)\')'
        + " trim(title), pcpsx_i_pel(), trim(c_flag), pc_get_lun()",
        "      IF (.true.) THEN",
        "        CALL smth",
        "      END IF",
        "      aaa: IF (.false.) THEN",
        "      ELSE IF (a) THEN",
        "      ELSE",
        "      END IF aaa",
        "      hey = 1",
        "    END SUBROUTINE bar",
        "    ABSTRACT INTERFACE",
        "    END INTERFACE",
        "  END MODULE foo",
    ]

    reader = fparser.common.readfortran.FortranStringReader(string)
    mode = fparser.common.sourceinfo.FortranFormat.from_mode("free")
    reader.set_format(mode)
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()
    caught = parser.block.tofortran().splitlines()
    assert caught[0][:13] == "!BEGINSOURCE "
    assert caught[1:] == expected


def test_module_procedure():
    """
    Tests a type that contains a procedure, and makes sure
    it has a module_procedures attribute
    """
    string = """
module foo
    type, public :: grid_type
   contains
        procedure :: get_tmask
   end type grid_type
end module foo
"""

    reader = fparser.common.readfortran.FortranStringReader(string)
    mode = fparser.common.sourceinfo.FortranFormat.from_mode("free")
    reader.set_format(mode)
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()

    # Fails if the class Type does not have a "module_procedures" attribute
    parser.analyze()


def test_f77():
    """
    Tests inherited from implementation code.
    """
    string = """\
      program foo
      a = 3
      end
      subroutine bar
      end
      pure function foo(a)
      end
      pure real*4 recursive function bar()
      end
"""
    expected = [
        "        PROGRAM foo",
        "          a = 3",
        "        END PROGRAM foo",
        "        SUBROUTINE bar()",
        "        END SUBROUTINE bar",
        "        pure FUNCTION foo(a)",
        "        END FUNCTION foo",
        "        pure recursive REAL*4 FUNCTION bar()",
        "        END FUNCTION bar",
    ]

    reader = fparser.common.readfortran.FortranStringReader(string)
    parser = fparser.one.parsefortran.FortranParser(reader)
    parser.parse()
    assert isinstance(parser.block, fparser.one.block_statements.BeginSource)
    caught = parser.block.tofortran().splitlines()
    assert caught[0][:25] == "      !      BEGINSOURCE "
    assert caught[1:] == expected



# Copyright (c) 2020-2025 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Test Fortran 2003 rule R470: this file tests the support for the
various forms of implicit loop within an array constructor.

TODO #257 - these tests need expanding.

Further tests are required here to fully cover the rule.

"""

import pytest
from fparser.common.readfortran import FortranStringReader
from fparser.two import Fortran2003


@pytest.mark.usefixtures("f2003_create")
def test_implicit_loop_constructor_no_parentheses():
    """Test that the parser does not match an implicit loop if the
    surrounding parentheses are missing (R470)."""
    fcode = "(/ival, ival=1, nval, istep/)"
    reader = FortranStringReader(fcode)
    ast = Fortran2003.Array_Constructor(reader)
    assert ast is None


@pytest.mark.usefixtures("f2003_create")
def test_no_implicit_loop_constructor_with_is_equal_to():
    """Test that the parser does not match an implicit loop if the
    array constructor contains '=='."""
    fcode = "(/lval1, ival1==ival2/)"
    reader = FortranStringReader(fcode)
    ast = Fortran2003.Array_Constructor(reader)
    # Should have an array constructor without an implicit loop.
    assert isinstance(ast, Fortran2003.Array_Constructor)
    assert isinstance(ast.children[1], Fortran2003.Ac_Value_List)


@pytest.mark.xfail(reason="#257 Constraint C497 is not checked.")
@pytest.mark.usefixtures("f2003_create")
def test_nested_implied_do():
    """Test C497 - the ac-do-variable of a nested implicit do shall not
    appear as the ac-do-variable of the containing implicit do."""
    reader = FortranStringReader("(/(JBODY,JBODY=1,SUM((/(JBODY,JBODY=1,4)/)))/)")
    ast = Fortran2003.Array_Constructor(reader)
    assert ast is None



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2020-2024, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ------------------------------------------------------------------------------
# Author: Joerg Henrichs, Bureau of Meteorology

"""This file contains an fparser script that parses Fortran files
and output the dependencies between these files suitable for a Makefile.

It assumes that the module name in the use statement corresponds to the
name of the file (adding one of .F90/.f90/.x90). Only files in the current
directory will be tested, so external dependencies will not be listed.

Usage:  create_dependencies.py  file1.f90 file2.F90 ...
"""

import os
import sys

from fparser.common.readfortran import FortranFileReader
from fparser.two.Fortran2003 import Use_Stmt
from fparser.two.parser import ParserFactory
from fparser.two.utils import walk


# -----------------------------------------------------------------------------
def usage():
    """This function prints the usage information and exits. It is called if
    incorrect input parameters are supplied.
    """
    print(f"{sys.argv[0]} file1 [file2...]")
    sys.exit(-1)


# -----------------------------------------------------------------------------
def get_root(file_name):
    """This function returns the 'root' of a file, which is the filename
    without any path component and without an extension. E.g.:
    `dir1/dir2/myfile.x90` would return `myfile`.

    :param str file_name: the name of the file.

    :returns: the root of the name of the file.
    :rtype: str

    """
    return os.path.splitext(os.path.basename(file_name))[0]


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    if len(sys.argv) < 2:
        usage()

    # Apparently pylint thinks that `all_files` is a constant, and complains
    # that it should be capitalised.
    # pylint: disable=invalid-name
    all_files = sys.argv[1:]

    # Create a mapping of filenames without path and extensions to the
    # full filename.
    lookup_files = {}
    for file in all_files:
        root = get_root(file)
        if root in lookup_files:
            print(
                f"The file '{file}' has the same root '{root}' as "
                f"the file '{lookup_files[root]}'."
            )
            sys.exit(-1)
        lookup_files[root] = file

    # Sort the input file names, so that they are output alphabetically
    all_files.sort()
    parser = ParserFactory().create(std="f2008")
    for filename in all_files:
        # Parse the current source file:
        try:
            reader = FortranFileReader(filename)
        except IOError:
            print(f"Could not open file '{filename}'.", file=sys.stderr)
            sys.exit(-1)

        parse_tree = parser(reader)

        # Collect all used modules in a list
        all_use = []
        for node in walk(parse_tree, Use_Stmt):
            # If you want to implement a specific naming convention,
            # you can modify the content of 'use_name' here. For example,
            # you could remove a '_mod' at the end if your file names do
            # not contains this.
            use_name = str(node.items[2])

            if use_name not in lookup_files:
                # Silently ignore modules we can't find, assuming
                # that they are system dependencies (e.g. MPI.mod)
                continue

            obj_dependency = os.path.splitext(lookup_files[use_name])[0] + ".o"
            # Only add dependencies that are not already in the list:
            if obj_dependency not in all_use:
                all_use.append(obj_dependency)

        # Now output all dependencies for this file (if any):
        if all_use:
            # Sort all module names on which this file depends,
            # so they are listed alphabetically
            all_use.sort()

            # Assemble the output in a list of strings. You start with
            # listing this file first:
            l_out = [os.path.splitext(filename)[0] + ".o: "]
            for dep_name in all_use:
                # Start the next line if we would exceed 80 characters:
                if len(l_out[-1]) + len(dep_name) > 80:
                    l_out.append("")

                l_out[-1] = l_out[-1] + dep_name
                # If this is not the last module, add a space:
                if dep_name != all_use[-1]:
                    l_out[-1] = l_out[-1] + " "

            print("\\\n\t".join(l_out))



# Copyright (c) 2021 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R726 : This file tests the support for a
Fortran character expression.

"""

import pytest
from fparser.two.Fortran2003 import (
    Default_Char_Expr,
    Char_Literal_Constant,
    Level_3_Expr,
)
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2003_create")
def test_simple_case():
    """Test that a simple expression gives the expected result."""
    result = Default_Char_Expr("'hello'")
    assert isinstance(result, Char_Literal_Constant)
    assert str(result) == "'hello'"
    assert repr(result).replace('u"', '"') == (
        "Char_Literal_Constant(\"'hello'\", None)"
    )


@pytest.mark.usefixtures("f2003_create")
def test_complicated_case():
    """Test that a more complicated expression gives the expected
    result.

    """
    result = Default_Char_Expr("'he'+'ll'//'o'")
    assert isinstance(result, Level_3_Expr)
    assert str(result) == "'he' + 'll' // 'o'"
    assert repr(result).replace("u'", "'").replace('u"', '"') == (
        "Level_3_Expr(Level_2_Expr(Char_Literal_Constant(\"'he'\", None), "
        "'+', Char_Literal_Constant(\"'ll'\", None)), '//', "
        "Char_Literal_Constant(\"'o'\", None))"
    )


@pytest.mark.parametrize(
    "string", ["1", "b'1010'", "o'7070'", "h'f0f0'", "1.0", "(1.0,1.0)", ".true."]
)
@pytest.mark.usefixtures("f2003_create")
def test_c707(string):
    """Check that invalid default literal constants do not match. Note,
    there are many other cases that are not currently checked in
    fparser.

    In theory -1 should become a signed_int_literal_constant and then
    fail to match. Similarly -1.0 should become a
    Signed_Real_Literal_Constant and then fail to match. However the
    "-" is treated as a unary expression so this never happens.

    """
    with pytest.raises(NoMatchError):
        _ = Default_Char_Expr(string)



# Copyright (c) 2020-2021 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Module containing py.test tests for the Io_Control_Spec class and
the list variant - r913.

"""

import pytest
from fparser.two import Fortran2003
from fparser.two.parser import ParserFactory

# this is required to setup the fortran2003 classes
_ = ParserFactory().create(std="f2003")

# Tests for the io_control_spec class.


def test_io_control_spec():
    """Test that we can construct an io-control-spec and that its str
    and repr properties are correct."""
    tcls = Fortran2003.Io_Control_Spec
    obj = tcls("end=123")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "END = 123"
    assert repr(obj) == "Io_Control_Spec('END', Label('123'))"


@pytest.mark.parametrize("string", ["6", "end 123"])
def test_io_ctrl_spec_named_only(string):
    """Check that io-control-spec only matches named entities."""
    tcls = Fortran2003.Io_Control_Spec
    with pytest.raises(Fortran2003.NoMatchError) as err:
        _ = tcls(string)
    assert "Io_Control_Spec: '{0}'".format(string) in str(err.value)


@pytest.mark.parametrize(
    "lhs, rhs",
    [
        ("unit", "num"),
        ("unit", "6"),
        ("fmt", "fmt_str"),
        ("fmt", "*"),
        ("fmt", "'(I3)'"),
        ("nml", "nml_name"),
        ("advance", "yes"),
        ("advance", "no"),
        ("asynchronous", "yes"),
        ("asynchronous", "no"),
        ("blank", "null"),
        ("blank", "zero"),
        ("decimal", "comma"),
        ("decimal", "point"),
        ("delim", "apostrophe"),
        ("end", "10"),
        ("eor", "10"),
        ("err", "10"),
        ("id", "var"),
        ("iomsg", "msg_var"),
        ("iostat", "var"),
        ("pad", "yes"),
        ("pad", "no"),
        ("pos", "posn"),
        ("pos", "10"),
        ("rec", "1"),
        ("round", "up"),
        ("round", "down"),
        ("round", "zero"),
        ("round", "nearest"),
        ("round", "compatible"),
        ("round", "processor_defined"),
        ("sign", "plus"),
        ("sign", "suppress"),
        ("sign", "processor_defined"),
        ("size", "var"),
    ],
)
def test_io_ctrl_spec_named_entities(lhs, rhs):
    """Check that all valid names are matched correctly."""
    obj = Fortran2003.Io_Control_Spec("=".join([lhs, rhs]))
    assert isinstance(obj, Fortran2003.Io_Control_Spec)
    assert obj.children[0] == lhs.upper()
    obj = Fortran2003.Io_Control_Spec("=".join([lhs.upper(), rhs]))
    assert isinstance(obj, Fortran2003.Io_Control_Spec)


@pytest.mark.parametrize(
    "lhs, rhs", [("asize", "var"), ("unit", "1.0"), ("size", "10")]
)
def test_io_ctrl_spec_invalid_values(lhs, rhs):
    """Check that io_ctrl_spec does not match if a keyword is unrecognised
    or the value is invalid."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Fortran2003.Io_Control_Spec("=".join([lhs, rhs]))


# Tests for the io_control_spec_list class.


def test_io_control_spec_list():
    """Test that we correctly parse and then generate various
    forms of IO-control specification lists (R913-list)."""
    tcls = Fortran2003.Io_Control_Spec_List
    obj = tcls("23, end=123")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "23, END = 123"
    assert (
        repr(obj) == "Io_Control_Spec_List(',', (Io_Control_Spec("
        "None, Int_Literal_Constant('23', None)), Io_Control_Spec('END', "
        "Label('123'))))"
    )

    obj = tcls("123")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123"

    obj = tcls("123,*")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123, *"
    assert repr(obj) == (
        "Io_Control_Spec_List(',', (Io_Control_Spec("
        "None, Int_Literal_Constant('123', None)), "
        "Io_Control_Spec(None, Format('*'))))"
    )

    obj = tcls("123,fmt=a")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123, FMT = a"
    assert repr(obj) == (
        "Io_Control_Spec_List(',', "
        "(Io_Control_Spec(None, "
        "Int_Literal_Constant('123', None)), "
        "Io_Control_Spec('FMT', Name('a'))))"
    )

    obj = tcls("123,nml=a")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123, NML = a"
    assert repr(obj) == (
        "Io_Control_Spec_List(',', "
        "(Io_Control_Spec(None, "
        "Int_Literal_Constant('123', None)), "
        "Io_Control_Spec('NML', Name('a'))))"
    )

    obj = tcls('123, "(I3)"')
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == '123, "(I3)"'

    obj = tcls("123,a")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123, a"

    # Unit named and not the first argument
    obj = tcls("fmt=b, unit=123")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "FMT = b, UNIT = 123"


def test_io_control_spec_list_3_or_more():
    """Test the matching when there are more than 2 entries in list - 3rd
    and subsequent entries must be named.
    """
    tcls = Fortran2003.Io_Control_Spec_List
    with pytest.raises(Fortran2003.NoMatchError) as err:
        _ = tcls("123,a,b")
    assert "Io_Control_Spec_List: '123,a,b'" in str(err.value)
    with pytest.raises(Fortran2003.NoMatchError) as err:
        _ = tcls("123,a,err=10,var")
    assert "Io_Control_Spec_List: '123,a,err=10,var'" in str(err.value)

    obj = tcls("123,a,err=10")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123, a, ERR = 10"

    obj = tcls("123,a,err=10,end=12")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "123, a, ERR = 10, END = 12"

    obj = tcls("unit=123,err=10,end=12")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "UNIT = 123, ERR = 10, END = 12"


def test_io_control_spec_list_invalid_first_entry():
    """Test that the special handling for the first item in the list
    rejects an invalid, unnamed entry."""
    tcls = Fortran2003.Io_Control_Spec_List
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("'name'")
    assert "Io_Control_Spec_List: ''name''" in str(err.value)
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("1.0")
    assert "Io_Control_Spec_List: '1.0'" in str(err.value)


def test_io_control_spec_list_invalid_2nd_entry():
    """Test that the special handling for the first two unnamed items in
    the list rejects an invalid, unnamed second entry."""
    tcls = Fortran2003.Io_Control_Spec_List
    # The second entry must be either a format specifier or a namelist-group-
    # name. We therefore supply a real literal as that doesn't match either.
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("unit_no, 1.0")
    assert "Io_Control_Spec_List: 'unit_no, 1.0'" in str(err.value)


def test_io_spec_list_c910():
    """Check that various constraints are applied in the matching. Currently
    only C910, C916-918 are implemented (#267)."""
    tcls = Fortran2003.Io_Control_Spec_List
    # C910 - must have a unit number
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls('fmt="some-fmt", end=10')
    assert "Io_Control_Spec_List: 'fmt=\"some-fmt\", end=10'" in str(err.value)
    # Again but when list contains only one entry
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls('fmt="some-fmt"')
    assert "Io_Control_Spec_List: 'fmt=\"some-fmt\"'" in str(err.value)


def test_io_spec_list_c916():
    """C916 - cannot have both a namelist and a format."""
    tcls = Fortran2003.Io_Control_Spec_List
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("123,nml=a,fmt=b")
    assert "Io_Control_Spec_List: '123,nml=a,fmt=b'" in str(err.value)
    # Check when we have an un-named second entry
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls('123, "(I3)", fmt=var')
    assert "Io_Control_Spec_List: '123, \"(I3)\", fmt=var'" in str(err.value)
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("123, nml_grp_name, fmt=var")
    assert "Io_Control_Spec_List: '123, nml_grp_name, fmt=var'" in str(err.value)


def test_io_spec_list_c917():
    """C917, if format is second item and is un-named then first item must be
    un-named unit number."""
    tcls = Fortran2003.Io_Control_Spec_List
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("unit=10, \"('a format')\"")
    assert "Io_Control_Spec_List: 'unit=10" in str(err.value)


def test_io_spec_list_c918():
    """C918, if namelist group name is second item and is un-named then first
    item must be un-named unit number."""
    tcls = Fortran2003.Io_Control_Spec_List
    with pytest.raises(Fortran2003.NoMatchError) as err:
        tcls("unit=10, nml_grp_name")
    assert "Io_Control_Spec_List: 'unit=10" in str(err.value)



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Coshape_Spec rule R511.a
"""

from fparser.common.splitline import string_replace_map
from fparser.two.Fortran2008.lower_cobound_r512 import Lower_Cobound
from fparser.two.Fortran2008.upper_cobound_r513 import Upper_Cobound
from fparser.two.utils import SeparatorBase


class Coshape_Spec(SeparatorBase):  # R511.a
    """
    coshape-spec is [ lower-cobound : ] upper-cobound

    """

    subclass_names = []
    use_names = ["Lower_Cobound", "Upper_Cobound"]

    @staticmethod
    def match(string):
        """
        Implements the matching for a coarray shape.

        :param str string: the string to match as shape.

        :return: `None` if there is no match, otherwise a 2-tuple with \
                 lower bound if given or `None`, and upper bound.
        :rtype: `NoneType` or \
            (`None`, :py:class:`fparser.two.Fortran2008.Upper_Cobound`) or \
            (:py:class:`fparser.two.Fortran2008.Lower_Cobound`, \
             :py:class:`fparser.two.Fortran2008.Upper_Cobound`)

        """
        line, repmap = string_replace_map(string)
        if ":" not in line:
            return (None, Upper_Cobound(string))
        lower, upper = line.split(":", 1)
        lower = lower.rstrip()
        upper = upper.lstrip()
        if not upper:
            return None
        if not lower:
            return None
        return (Lower_Cobound(repmap(lower)), Upper_Cobound(repmap(upper)))

    def tostr(self):
        """
        Converts the Shape specification to string.

        :return: the shape specification as string.
        :rtype: str

        """
        if self.items[0] is None:
            return str(self.items[1])
        return SeparatorBase.tostr(self)



# Copyright (c) 2021 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R705 : This file tests support for the
Add_Operand class.

"""

import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2003 import Add_Operand, Level_2_Expr
from fparser.two.parser import ParserFactory

# This is required to setup the fortran2003 classes (when matching
# with Add_Operand)
_ = ParserFactory().create(std="f2003")


@pytest.mark.parametrize(
    "string,str_repr",
    [
        ("a ** b", "Mult_Operand(Name('a'), '**', Name('b'))"),
        (".DEFINED. a", "Level_1_Expr('.DEFINED.', Name('a'))"),
        ("1.0", "Real_Literal_Constant('1.0', None)"),
        (
            "(a + b)",
            "Parenthesis('(', Level_2_Expr(Name('a'), '+', " "Name('b')), ')')",
        ),
    ],
)
def test_mult(string, str_repr):
    """Test for a successful match with a valid mult-operand string"""
    result = Add_Operand(string)
    assert str(result) == string
    assert repr(result) == str_repr


def test_mult_fail():
    """Test for a failure to match an invalid multi-operand string"""
    with pytest.raises(NoMatchError) as info:
        Add_Operand("a .and. b")
    assert "Add_Operand: 'a .and. b'" in str(info.value)


@pytest.mark.parametrize(
    "string,str_repr",
    [
        ("a * b", "Add_Operand(Name('a'), '*', Name('b'))"),
        (
            "a * 1.0E-3",
            "Add_Operand(Name('a'), '*', " "Real_Literal_Constant('1.0E-3', None))",
        ),
        ("a / b", "Add_Operand(Name('a'), '/', Name('b'))"),
        (
            "a * b * c",
            "Add_Operand(Add_Operand(Name('a'), '*', Name('b')), " "'*', Name('c'))",
        ),
        (
            "a * b / c",
            "Add_Operand(Add_Operand(Name('a'), '*', Name('b')), " "'/', Name('c'))",
        ),
    ],
)
def test_add_operand(string, str_repr):
    """Test for a successful match with a valid add_operand"""
    result = Add_Operand(string)
    assert str(result) == string
    assert repr(result) == str_repr


@pytest.mark.parametrize("string", ["a + b", "_ * b", "a * _", "a * b + c"])
def test_add_operand_fail(string):
    """Test for a failure to match an invalid add-operand string"""
    with pytest.raises(NoMatchError) as info:
        Add_Operand(string)
    assert "Add_Operand: '{0}'".format(string) in str(info.value)



# Copyright (c) 2017-2022 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Test the various utility functions

"""

import pytest

from fparser.common.utils import split_comma, ParseError


def test_split_comma():
    """Test the split_comma() function"""
    items = split_comma("hello, goodbye")
    print(items)
    assert items[0] == "hello"
    assert items[1] == "goodbye"
    # With trailing and leading white space
    items = split_comma("  hello, goodbye   ")
    print(items)
    assert items[0] == "hello"
    assert items[1] == "goodbye"
    items = split_comma("  ")
    assert not items


def test_split_comma_exceptions():
    """Test that we raise the expected exceptions if we don't supply
    the brackets in the right form"""
    with pytest.raises(ParseError) as excinfo:
        _ = split_comma("one, two", brackets="()")
    assert "brackets must be a tuple" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = split_comma("one, two", brackets=("()",))
    assert "brackets tuple must contain just two items" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = split_comma("one, two", brackets=("(", "(", "("))
    assert "brackets tuple must contain just two items" in str(excinfo.value)


def test_split_bracketed_list():
    """Test the splitting of a list bracketed with parentheses"""
    items = split_comma("(well(1), this(is), it)", brackets=("(", ")"))
    print(items)
    assert items[0] == "well(1)"
    assert items[1] == "this(is)"
    # With superfluous white space
    items = split_comma("  (  well(1), this(is), it  )  ", brackets=("(", ")"))
    print(items)
    assert items[0] == "well(1)"
    assert items[1] == "this(is)"
    assert items[2] == "it"
    # Square brackets
    items = split_comma("[well(1), this(is), it]", brackets=("[", "]"))
    print(items)
    assert items[0] == "well(1)"
    assert items[1] == "this(is)"
    assert items[2] == "it"
    # Mis-matched brackets
    items = split_comma("[well(1), this(is), it)", brackets=("[", "]"))
    assert not items


def test_extract_bracketed_list():
    """Test the extraction and parsing of a list within parentheses within
    a larger string"""
    from fparser.common.utils import extract_bracketed_list_items

    items = extract_bracketed_list_items("hello (this, is, a) test")
    assert items[0] == "this"
    assert items[1] == "is"
    assert items[2] == "a"


def test_extract_bracketed_list_err():
    """Test that we get the expected errors if the string passed into
    extract_bracketed_list_items() does not have the correct format"""
    from fparser.common.utils import extract_bracketed_list_items

    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello (this, is, wrong(")
    assert "more than one opening/closing parenthesis found" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello )this, is, wrong)")
    assert "more than one opening/closing parenthesis found" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello (this, is, wrong) (too)")
    assert "more than one opening/closing parenthesis found" in str(excinfo.value)
    with pytest.raises(ParseError) as excinfo:
        _ = extract_bracketed_list_items("hello )this, is, wrong( too")
    assert "failed to find expression within parentheses in" in str(excinfo.value)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2024, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ------------------------------------------------------------------------------
# Author: Joerg Henrichs, Bureau of Meteorology

"""This file contains an fparser script that splits a source file that contains
several modules and/or subroutines into individual files. If a file that needs
to be created already exists, the script will abort.

Usage:  split_file.py  file.f90
"""

import os
import subprocess
import sys

from fparser.common.readfortran import FortranFileReader
from fparser.two import C99Preprocessor, Fortran2003
from fparser.two.parser import ParserFactory
from fparser.two.utils import walk


# -----------------------------------------------------------------------------
def usage():
    """This function prints the usage information and exits. It is called if
    incorrect input parameters are supplied.
    """
    print(f"{sys.argv[0]} file1")
    sys.exit(-1)


# -----------------------------------------------------------------------------
def create_makefile(main_name, all_objs, all_filenames):
    """This function creates a Makefile to compile (and if a main_name is
    given to link) the binaries. It uses the create_dependencies.py script
    to add the actual dependencies between the modules to the makefile.

    :param str main_name: name of the program, or "" if no program was
        contained in the original file.
    :param List[str] all_objs: the list of all objects to compile.
    :param List[str] all_filenames: the list of all file names, used to
        call the `create_dependency.py` script

    """
    file_path = os.path.dirname(os.path.realpath(__file__))
    completed = subprocess.run(
        [f"{file_path}/create_dependencies.py"] + all_filenames,
        capture_output=True,
        check=True,
    )

    # Query some environment flags which will define the
    # defaults in the Makefile
    f90 = os.getenv("F90", "gfortran")
    f90flags = os.getenv("F90FLAGS", "-g -O0")
    ldflags = os.getenv("LDFLAGS", "")

    makefile = "Makefile"
    if os.path.exists(makefile):
        print(f"The file '{makefile}' already exists - aborting.")
        sys.exit(-1)

    # Now create a makefile
    if main_name:
        default_target = (
            f"default: {main_name}\n\n"
            f"{main_name}: $(OBJS)\n"
            f"\t$(F90) $(F90FLAGS) $(OBJS) -o {main_name} "
            f"$(LDFLAGS)"
        )
        clean_actions = f"\trm -f {main_name} $(OBJS) *.mod"
    else:
        default_target = "default: $(OBJS)"
        clean_actions = "\trm -f $(OBJS) *.mod"

    with open(makefile, mode="w", encoding="utf-8") as f_out:
        f_out.write(f"""
F90 ?= {f90}
# We have to enforce this setting, since using ?= will not
# change the value of CPP, which will then be using `cc -E`, which
# in turn does not handle indented preprocessor lines.
CPP = cpp
# Don't use traditional, it does also not accept indented preprocessor lines
CPPFLAGS ?= -P
F90FLAGS ?= {f90flags}
LDFLAGS ?= {ldflags}

OBJS={' '.join(all_objs)}

{default_target}

# Dependencies
# ============
{completed.stdout.decode("utf-8")}

# Preprocessing (required since some compilers do not
# handle the indented preprocessor directives of fparser)
# =======================================================
%.f90: %.F90
\t$(CPP) $(CPPFLAGS) $< > $@

# Compilation rules
# =================
%.o: %.f90
\t$(F90) $(F90FLAGS) -c $<

# Cleanup
# =======
clean:
{clean_actions}
""")


# -----------------------------------------------------------------------------
def main():
    """The main program. Having this as a dedicated function solves many
    pylint complains.

    """
    if len(sys.argv) != 2:
        usage()

    try:
        reader = FortranFileReader(sys.argv[1])
    except IOError:
        print(f"Could not open file '{sys.argv[1]}'.", file=sys.stderr)
        sys.exit(-1)

    Parser = ParserFactory().create(std="f2008")
    parse_tree = Parser(reader)

    # Get a list of all preprocessor classes (which will determine
    # the extension to be used: F90/f90
    all_cpp_classes = tuple(
        getattr(C99Preprocessor, i) for i in C99Preprocessor.CPP_CLASS_NAMES
    )
    all_filenames = []
    all_objs = []
    main_name = ""
    for unit in parse_tree.children:
        # For Module, Subroutine_Subprogram, Function_Subprogram, Main_Program
        # we always get the name using this:
        unit_name = unit.children[0].items[1].string.lower()

        # Save the name of the main program if we have one, which will be
        # used in the Makefile created later
        if isinstance(unit, Fortran2003.Main_Program):
            main_name = unit_name
        if any(walk(unit, all_cpp_classes)):
            ext = ".F90"
        else:
            ext = ".f90"

        filename = f"{unit_name}{ext}"
        if os.path.exists(filename):
            print(f"The file '{filename}' already exists - aborting.")
            sys.exit(-1)
        with open(filename, mode="w", encoding="utf-8") as f_out:
            f_out.write(str(unit))
            f_out.write("\n")
        all_filenames.append(filename)
        all_objs.append(f"{unit_name}.o")

    # Now create the makefile
    create_makefile(main_name, all_objs, all_filenames)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    # Using a function here avoids pylint warnings
    main()



# Copyright (c) 2020-2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R501.

type-declaration-stmt is declaration-type-spec [ [ , attr-spec ] ... :: ]
                         entity-decl-list

"""

import pytest
from fparser.two.Fortran2003 import Intrinsic_Function_Reference
from fparser.two.Fortran2008 import Type_Declaration_Stmt
from fparser.two.symbol_table import SYMBOL_TABLES
from fparser.two.utils import walk
from fparser.api import get_reader


@pytest.mark.usefixtures("fake_symbol_table")
def test_type_declaration_stmt():  # R501
    """
    Tests copied from Fortran 2003.

    """
    tcls = Type_Declaration_Stmt
    obj = tcls("integer a")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "INTEGER :: a"
    assert (
        repr(obj) == "Type_Declaration_Stmt(Intrinsic_Type_Spec('INTEGER', "
        "None), None, Entity_Decl_List(',', (Entity_Decl(Name('a'), None, "
        "None, None),)))"
    )

    obj = tcls("integer ,dimension(2):: b*3")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "INTEGER, DIMENSION(2) :: b*3"

    obj = tcls("real c")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "REAL :: c"
    assert (
        repr(obj) == "Type_Declaration_Stmt(Intrinsic_Type_Spec('REAL', None), None, "
        "Entity_Decl_List(',', (Entity_Decl(Name('c'), None, None, "
        "None),)))"
    )

    obj = tcls("REAL D( LDA, * ), E( LDB, * )")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "REAL :: D(LDA, *), E(LDB, *)"

    obj = tcls("DOUBLE PRECISION   ALPHA, BETA")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "DOUBLE PRECISION :: ALPHA, BETA"

    obj = tcls("logical,parameter:: T=.true.")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "LOGICAL, PARAMETER :: T = .TRUE."

    obj = tcls("character(n),private:: x(n)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CHARACTER(LEN = n), PRIVATE :: x(n)"

    obj = tcls("character(lenmax),private:: y(n)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CHARACTER(LEN = lenmax), PRIVATE :: y(n)"


def test_shadowed_intrinsic(f2008_parser):
    """Check that a locally-defined symbol that shadows (overwrites) a
    Fortran intrinsic is correctly identified."""
    tree = f2008_parser(get_reader("""\
module my_mod
  use some_mod
  real :: dot_product(2,2)
contains
  subroutine my_sub()
    real :: result
    result = dot_product(1,1)
  end subroutine my_sub
end module my_mod
    """))
    tables = SYMBOL_TABLES
    # We should not have an intrinsic-function reference in the parse tree
    assert not walk(tree, Intrinsic_Function_Reference)
    table = tables.lookup("my_mod")
    sym = table.children[0].lookup("dot_product")
    assert sym.primitive_type == "real"



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Loop_Control rule R818
"""

from fparser.two.Fortran2003 import Loop_Control as Loop_Control_2003, Forall_Header


class Loop_Control(Loop_Control_2003):  # R818
    """Fortran 2008 rule R818

    loop-control is [ , ] do-variable = scalar-int-expr , scalar-int-expr
                       [ , scalar-int-expr ]
                    or [ , ] WHILE ( scalar-logical-expr )
                    or [ , ] CONCURRENT forall-header

    Extends the Fortran2003 rule R830 with the additional CONCURRENT clause.

    The F2003 Loop_Control class would be better and more extensible
    if it called 2 classes, one for each of the above
    expressions. This would then affect the implementation of this
    class. Something like the suggestion below. However, this would
    result in a different fparser tree, see issue #416.

    F2003 - While_Loop_Cntl: scalar-logical-expression, delim
    F2003 - Counter_Loop_Cntl: var, lower, upper, [step], delim
    F2008 - Concurrent_Loop_Cntl: conc_expr, delim
    F2018 - Concurrent_Loop_Cntl: conc_expr, local_x, delim

    """

    subclass_names = []
    # This class' match method makes use of the Fortran2003 match
    # method so 'use_names' should include any classes used within
    # there as well as any used here.
    use_names = Loop_Control_2003.use_names[:]
    use_names.append("Forall_Header")

    @staticmethod
    def match(string):
        """Attempts to match the supplied text with this rule.

        :param str string: Fortran code to check for a match.

        :returns: None if there is no match, a tuple with the first \
            entry providing the result of matching the 'WHILE' part of \
            the rule if there is a match, the second entry providing \
            the result of matching the 'COUNTER' part of the rule if \
            there is a match, the third entry indicating whether \
            there is an optional preceding ',' and the fourth entry \
            providing the result of matching the 'CONCURRENT' part of \
            the rule if there is a match.

        :rtype: Optional[Tuple[ \
            Optional[ \
                :py:class:`fparser.two.Fortran2003.Scalar_Logical_Expr`], \
            Optional[Tuple[ \
                :py:class:`fparser.two.Fortran2003.Do_Variable`, List[str]]], \
            Optional[str], \
            Optional[:py:class:`fparser.two.Fortran2003.Forall_Header`]]]

        """
        # Fortran2003 matches all but CONCURRENT so try this first
        result = Loop_Control_2003.match(string)
        if result:
            return result + (None,)
        # Try to match with CONCURRENT
        line = string.lstrip()
        optional_delim = None
        if line.startswith(","):
            line = line[1:].lstrip()
            optional_delim = ","
        if line[:10].upper() != "CONCURRENT":
            return None
        return (None, None, optional_delim, Forall_Header(line[10:].lstrip().rstrip()))

    def tostr(self):
        """
        :returns: the Fortran representation of this object.
        :rtype: str
        """
        if self.items[0] or self.items[1]:
            # Use the F2003 tostr() implementation
            return Loop_Control_2003.tostr(self)
        # Return loop control construct containing "CONCURRENT" clause
        loopctrl = f"CONCURRENT {self.items[3]}"
        # Add optional delimiter to loop control construct if present
        if self.items[2]:
            loopctrl = f"{self.items[2]} {loopctrl}"
        return loopctrl
